==> intel-oneapi-mkl-2022.1.0_223-1-aarch64.pkg.tar.xz: c_offload/fftw3x/source/sp_plan_many_dft_1d_async.c <==
/*******************************************************************************
* Copyright 2020-2022 Intel Corporation.
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
! Content:
!       Example of using fftwf_plan_many_dft function on a
!       (GPU) device using the OpenMP target (offload) interface..
!
!****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>
#include "fftw/fftw3.h"
#include "fftw/offload/fftw3_omp_offload.h"

static void init(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H);
static int verify(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H);

int main(void)
{
    /*
     * In this example we perform M in-place 1D FFT on the data
     * contained in a larger array.  The sizes of the FFT are defined
     * by N, the embedding array has dimensions defined by
     * EN and EM, not necessarily in this order.
     */

    /* Sizes of 1D transform and the number of them */
    int N = 64;
    int M = 7;

    /* Sizes of embedding array, stride and distance, to be defined */
    int EN, EM, stride, dist;

    /* Arbitrary harmonic used to verify FFT */
    int H = -N/2;

    /* FFTW plan handle */
    fftwf_plan plan = 0;

    /* Pointer to input/output data */
    fftwf_complex *x = 0;

    /* Execution status */
    int status = 0;

    const int devNum = 0;

    printf("Example sp_plan_many_dft_1d_async\n");
    printf("Forward multiple 1D complex in-place FFT\n");
    printf("Configuration parameters:\n");
    printf(" N  = {%d}\n", N);
    printf(" M  = %d\n", M);
    printf(" H  = {%d}\n", H);

    printf("Define data layout for PARALLEL transforms\n");
    /*
     * Leading dimension is N (parallel transforms): embedding array
     * has dimensions (EM,EN).
     */
    stride = 1;
    EN   = N * stride; // EN same as N
    dist = EN;
    EM     = M;
    printf(" EM=%i, dist=%i, EN={%i}, stride=%i\n",
           EM,dist,EN,stride);


    printf("Allocate x(%i)\n", EM*dist );
    x  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * EM*dist );
    if (0 == x) goto failed;

    printf("Initialize input for forward transform\n");
    init(x, N, M, EN, stride, dist, H);

    printf("Create FFTW plan for forward transform\n");
#pragma omp target data map(tofrom:x[0:EM*dist]) device(devNum)
    {
#pragma omp target variant dispatch use_device_ptr(x) device(devNum)
    {
        plan = fftwf_plan_many_dft(1, &N, M,
                                   x, &EN, stride, dist,
                                   x, &EN, stride, dist,
                                   FFTW_FORWARD, FFTW_ESTIMATE);
    }
    if(plan == 0) printf("Call to fftwf_plan_many_dft has failed\n");

    printf("Compute forward FFT\n");
#pragma omp target variant dispatch device(devNum) nowait
    {
        fftwf_execute(plan);
    }
#pragma omp taskwait
    }

    printf("Verify the result of forward FFT\n");
    status = verify(x, N, M, EN, stride, dist, H);
    if (0 != status) goto failed;

    printf("Destroy FFTW plan\n");
    fftwf_destroy_plan(plan);

    printf("Free data array\n");
    fftwf_free(x);


    printf("\nDefine data layout for VECTOR transforms\n");
    /*
     * Leading dimension is M (vector transforms): embedding array
     * has dimensions (EN,EM).
     */
    dist   = 1;
    EM     = M * dist;
    stride = EM;
    EN = N;

    printf(" EN={%i}, stride=%i, EM=%i, dist=%i\n",
           EN,stride,EM,dist);

    printf("Allocate x(%i)\n", EN*stride );
    x  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * EN*stride );
    if (0 == x) goto failed;

    printf("Initialize input for forward transform\n");
    init(x, N, M, EN, EM, dist, H);

    printf("Create FFTW plan for forward transform\n");
#pragma omp target data map(tofrom:x[0:EN*stride]) device(devNum)
    {
#pragma omp target variant dispatch use_device_ptr(x) device(devNum)
    {
        plan = fftwf_plan_many_dft(1, &N, M,
                                   x, &EN, stride, dist,
                                   x, &EN, stride, dist,
                                   FFTW_FORWARD, FFTW_ESTIMATE);
    }
    if(plan == 0) printf("Call to fftwf_plan_many_dft has failed\n");

    printf("Compute forward FFT\n");
#pragma omp target variant dispatch device(devNum) nowait
    {
        fftwf_execute(plan);
    }
#pragma omp taskwait
    }

    printf("Verify the result of forward FFT\n");
    status = verify(x, N, M, EN, EM, dist, H);
    if (0 != status) goto failed;

 cleanup:

    printf("Destroy FFTW plan\n");
    fftwf_destroy_plan(plan);

    printf("Free data array\n");
    fftwf_free(x);

    printf("TEST %s\n",0==status ? "PASSED" : "FAILED");
    return status;

 failed:
    printf(" ERROR\n");
    status = 1;
    goto cleanup;
}

/* Compute (K*L)%M accurately */
static float moda(int K, int L, int M)
{
    return (float)(((long long)K * L) % M);
}

/* Initialize arrays x with harmonic H */
static void init(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H)
{
    float TWOPI = 6.2831853071795864769f, phase;
    int n1, m, N1, S1, SM, H1, index;

    SM = dist;
    S1 = stride; N1 = N; H1 = H;

    for (m = 0; m < M; m++)
    {
        for(n1 = 0; n1 < N; n1++)
        {
            phase  = moda(n1,H1,N1) / N1;
            index = n1*S1 + m*SM;
            x[index][0] = cosf( TWOPI * phase ) / (N1);
            x[index][1] = sinf( TWOPI * phase ) / (N1);
        }
    }

}

/* Verify that x has unit peak at H */
static int verify(fftwf_complex *x,
                  int N, int M, int EN, int stride, int dist, int H)
{
    float err, errthr, maxerr;
    int n1, m, N1, S1, SM, H1, index;

    SM = dist;
    S1 = stride; N1 = N; H1 = H;

    /*
     * Note, this simple error bound doesn't take into account error of
     * input data
     */
    errthr = 5.0f * logf( (float)N1 ) / logf(2.0f) * FLT_EPSILON;
    printf(" Check if err is below errthr %.3g\n", errthr);

    maxerr = 0;
    for (m = 0; m < M; m++)
    {
        for (n1 = 0; n1 < N1; n1++)
        {
            float re_exp = 0.0, im_exp = 0.0, re_got, im_got;

            if ((n1-H1)%N1==0) re_exp = 1;

            index = n1*S1 + m*SM;
            re_got = x[index][0];
            im_got = x[index][1];
            err  = fabsf(re_got - re_exp) + fabsf(im_got - im_exp);
            if (err > maxerr) maxerr = err;
            if (!(err < errthr))
            {
                printf(" x(n1=%i,m=%i): ",n1,m);
                printf(" expected (%.7g,%.7g), ",re_exp,im_exp);
                printf(" got (%.7g,%.7g), ",re_got,im_got);
                printf(" err %.3g\n", err);
                printf(" Verification FAILED\n");
                return 1;
            }
        }
    }
    printf(" Verified, maximum error was %.3g\n", maxerr);
    return 0;
}

==> intel-oneapi-mkl-2023.2.0_49495-2-aarch64.pkg.tar.xz: c_offload/fftw3x/source/sp_plan_many_dft_1d_async.c <==
/*******************************************************************************
* Copyright 2020-2022 Intel Corporation.
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
! Content:
!       Example of using fftwf_plan_many_dft function on a
!       (GPU) device using the OpenMP target (offload) interface..
!
!****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>
#include "fftw/fftw3.h"
#include "fftw/offload/fftw3_omp_offload.h"

static void init(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H);
static int verify(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H);

int main(void)
{
    /*
     * In this example we perform M in-place 1D FFT on the data
     * contained in a larger array.  The sizes of the FFT are defined
     * by N, the embedding array has dimensions defined by
     * EN and EM, not necessarily in this order.
     */

    /* Sizes of 1D transform and the number of them */
    int N = 64;
    int M = 7;

    /* Sizes of embedding array, stride and distance, to be defined */
    int EN, EM, stride, dist;

    /* Arbitrary harmonic used to verify FFT */
    int H = -N/2;

    /* FFTW plan handle */
    fftwf_plan plan = 0;

    /* Pointer to input/output data */
    fftwf_complex *x = 0;

    /* Execution status */
    int status = 0;

    const int devNum = 0;

    printf("Example sp_plan_many_dft_1d_async\n");
    printf("Forward multiple 1D complex in-place FFT\n");
    printf("Configuration parameters:\n");
    printf(" N  = {%d}\n", N);
    printf(" M  = %d\n", M);
    printf(" H  = {%d}\n", H);

    printf("Define data layout for PARALLEL transforms\n");
    /*
     * Leading dimension is N (parallel transforms): embedding array
     * has dimensions (EM,EN).
     */
    stride = 1;
    EN   = N * stride; // EN same as N
    dist = EN;
    EM     = M;
    printf(" EM=%i, dist=%i, EN={%i}, stride=%i\n",
           EM,dist,EN,stride);


    printf("Allocate x(%i)\n", EM*dist );
    x  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * EM*dist );
    if (0 == x) goto failed;

    printf("Initialize input for forward transform\n");
    init(x, N, M, EN, stride, dist, H);

    printf("Create FFTW plan for forward transform\n");
#pragma omp target data map(tofrom:x[0:EM*dist]) device(devNum)
    {
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch use_device_ptr(x) device(devNum)
#endif
    plan = fftwf_plan_many_dft(1, &N, M,
                               x, &EN, stride, dist,
                               x, &EN, stride, dist,
                               FFTW_FORWARD, FFTW_ESTIMATE);
    if(plan == 0) printf("Call to fftwf_plan_many_dft has failed\n");

    printf("Compute forward FFT\n");
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum) nowait
#else
#pragma omp target variant dispatch device(devNum) nowait
#endif
    fftwf_execute(plan);
#pragma omp taskwait
    }

    printf("Verify the result of forward FFT\n");
    status = verify(x, N, M, EN, stride, dist, H);
    if (0 != status) goto failed;

    printf("Destroy FFTW plan\n");
    fftwf_destroy_plan(plan);

    printf("Free data array\n");
    fftwf_free(x);


    printf("\nDefine data layout for VECTOR transforms\n");
    /*
     * Leading dimension is M (vector transforms): embedding array
     * has dimensions (EN,EM).
     */
    dist   = 1;
    EM     = M * dist;
    stride = EM;
    EN = N;

    printf(" EN={%i}, stride=%i, EM=%i, dist=%i\n",
           EN,stride,EM,dist);

    printf("Allocate x(%i)\n", EN*stride );
    x  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * EN*stride );
    if (0 == x) goto failed;

    printf("Initialize input for forward transform\n");
    init(x, N, M, EN, EM, dist, H);

    printf("Create FFTW plan for forward transform\n");
#pragma omp target data map(tofrom:x[0:EN*stride]) device(devNum)
    {
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch use_device_ptr(x) device(devNum)
#endif
    plan = fftwf_plan_many_dft(1, &N, M,
                               x, &EN, stride, dist,
                               x, &EN, stride, dist,
                               FFTW_FORWARD, FFTW_ESTIMATE);
    if(plan == 0) printf("Call to fftwf_plan_many_dft has failed\n");

    printf("Compute forward FFT\n");
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum) nowait
#else
#pragma omp target variant dispatch device(devNum) nowait
#endif
    fftwf_execute(plan);
#pragma omp taskwait
    }

    printf("Verify the result of forward FFT\n");
    status = verify(x, N, M, EN, EM, dist, H);
    if (0 != status) goto failed;

 cleanup:

    printf("Destroy FFTW plan\n");
    fftwf_destroy_plan(plan);

    printf("Free data array\n");
    fftwf_free(x);

    printf("TEST %s\n",0==status ? "PASSED" : "FAILED");
    return status;

 failed:
    printf(" ERROR\n");
    status = 1;
    goto cleanup;
}

/* Compute (K*L)%M accurately */
static float moda(int K, int L, int M)
{
    return (float)(((long long)K * L) % M);
}

/* Initialize arrays x with harmonic H */
static void init(fftwf_complex *x,
                 int N, int M, int EN, int stride, int dist, int H)
{
    float TWOPI = 6.2831853071795864769f, phase;
    int n1, m, N1, S1, SM, H1, index;

    SM = dist;
    S1 = stride; N1 = N; H1 = H;

    for (m = 0; m < M; m++)
    {
        for(n1 = 0; n1 < N; n1++)
        {
            phase  = moda(n1,H1,N1) / N1;
            index = n1*S1 + m*SM;
            x[index][0] = cosf( TWOPI * phase ) / (N1);
            x[index][1] = sinf( TWOPI * phase ) / (N1);
        }
    }

}

/* Verify that x has unit peak at H */
static int verify(fftwf_complex *x,
                  int N, int M, int EN, int stride, int dist, int H)
{
    float err, errthr, maxerr;
    int n1, m, N1, S1, SM, H1, index;

    SM = dist;
    S1 = stride; N1 = N; H1 = H;

    /*
     * Note, this simple error bound doesn't take into account error of
     * input data
     */
    errthr = 5.0f * logf( (float)N1 ) / logf(2.0f) * FLT_EPSILON;
    printf(" Check if err is below errthr %.3g\n", errthr);

    maxerr = 0;
    for (m = 0; m < M; m++)
    {
        for (n1 = 0; n1 < N1; n1++)
        {
            float re_exp = 0.0, im_exp = 0.0, re_got, im_got;

            if ((n1-H1)%N1==0) re_exp = 1;

            index = n1*S1 + m*SM;
            re_got = x[index][0];
            im_got = x[index][1];
            err  = fabsf(re_got - re_exp) + fabsf(im_got - im_exp);
            if (err > maxerr) maxerr = err;
            if (!(err < errthr))
            {
                printf(" x(n1=%i,m=%i): ",n1,m);
                printf(" expected (%.7g,%.7g), ",re_exp,im_exp);
                printf(" got (%.7g,%.7g), ",re_got,im_got);
                printf(" err %.3g\n", err);
                printf(" Verification FAILED\n");
                return 1;
            }
        }
    }
    printf(" Verified, maximum error was %.3g\n", maxerr);
    return 0;
}
